from .list import get_list_of_items


def get_api_second_list(api_information_url):
    counter = 0
    first_list_found = False
    second_list = []

    while len(second_list) == 0 or not first_list_found:
        second_list = get_list_of_items(api_information_url, counter)
        if len(second_list) > 0 and not first_list_found:
            second_list = []
            first_list_found = True

        counter += 1

    return second_list



from .list import get_list_of_items


def get_api_third_list(api_information_url):
    counter = 0
    first_list_found = False
    second_list_found = False
    third_list = []

    while len(third_list) == 0 or not first_list_found or not second_list_found:
        third_list = get_list_of_items(api_information_url, counter)
        if len(third_list) > 0 and not first_list_found:
            third_list = []
            first_list_found = True
        elif len(third_list) > 0 and not second_list_found:
            third_list = []
            second_list_found = True

        counter += 1

    return third_list



# This is the base URL from which we are exploring docs
SHOPIFY_BASE_URL = "https://shopify.dev"

# This is the API URL that gets called internally to figure out what links to render in the navbar
SHOPIFY_NAV_API_URL = "https://shopify.dev/docs/api/admin-graphql/latest/nav"

# Available languages to view code examples in
EXAMPLE_LANGUAGES = ["GQL", "cURL", "Remix", "Node.js", "Ruby"]

# The language to be getting code examples in
EXAMPLE_LANGUAGE = EXAMPLE_LANGUAGES[4] # "Ruby"

# The Llama model to request from ollama
OLLAMA_MODEL = 'llama3.2'



from shopagent.models import Object

from .code_examples import get_api_examples
from .description import get_api_description
from .fields_and_connections import get_api_fields_or_connections
from .returns import get_api_returns


def get_object_info(label, api_information_url):
    name = label
    description = get_api_description(api_information_url)
    fields_and_connections = get_api_fields_or_connections(api_information_url, True)
    returns = get_api_returns(api_information_url, False, len(fields_and_connections) > 0)
    examples = get_api_examples(api_information_url)

    obj = Object(
        name=name,
        description=description.strip(),
        fields_and_connections=fields_and_connections,
        examples=examples,
    )

    return obj



from .first_list import get_first_list
from .second_list import get_second_list


def get_possible_second_list(api_information_url, first_list_found):
    if first_list_found:
        return get_second_list(api_information_url)
    return get_first_list(api_information_url)



from shopagent.models.api import ShopifyAPI, ShopifyAPIBreakdown
from shopagent.nav import complete_nav_url, get_top_level_operations

from .mutation import get_mutation_info
from .object import get_object_info
from .query import get_query_info


def get_api_information(shopify_nav):
    categories = []

    encountered_objects = 0
    total_objects = sum([sum(len(child.children) for child in nav.children) for nav in shopify_nav.navItems])

    for i, nav in enumerate(shopify_nav.navItems):
        print(f"Working on {i+1} nav item of {len(shopify_nav.navItems)}")
        name = nav.label
        mutations = []
        objects = []
        queries = []

        for j, child in enumerate(nav.children):
            print(f"Working on {j+1} nav item child of {len(nav.children)}")
            if child.label == "Mutations":
                for k, mutation_child in enumerate(child.children):
                    print(f"Working on {k+1} mutation nested child of {len(child.children)} <{encountered_objects}/{total_objects}>")
                    mutations += [get_mutation_info(mutation_child.label, complete_nav_url(mutation_child.href))]
                    encountered_objects += 1
            elif child.label == "Objects":
                for k, object_child in enumerate(child.children):
                    print(f"Working on {k+1} object nested child of {len(child.children)} <{encountered_objects}/{total_objects}>")
                    objects += [get_object_info(object_child.label, complete_nav_url(object_child.href))]
                    encountered_objects += 1
            elif child.label == "Queries":
                for k, query_child in enumerate(child.children):
                    print(f"Working on {k+1} query nested child of {len(child.children)} <{encountered_objects}/{total_objects}>")
                    queries += [get_query_info(query_child.label, complete_nav_url(query_child.href))]
                    encountered_objects += 1

        categories += [ShopifyAPIBreakdown(
            name=name,
            queries=queries,
            mutations=mutations,
            objects=objects
        )]    

    return ShopifyAPI(categories=categories)


def get_data():
    shopify_nav = get_top_level_operations()
    api_information = get_api_information(shopify_nav)

    with open("shopify_api.json", "w") as shopify_api_file:
        json.dump(api_information.model_dump(), shopify_api_file)



from shopagent.lsd import run_lsd

from .first_list import get_first_list
from .second_list import get_api_second_list
from .third_list import get_api_third_list


HAS_API_RETURNS_SQL = """
returns_anchor <| a#possiblereturns |

FROM {api_information_url}
|> GROUP BY returns_anchor
|> SELECT returns_anchor
"""


def get_api_returns(api_information_url, start_at_second, prior_list_found):
    global HAS_API_RETURNS_SQL

    results = run_lsd(HAS_API_RETURNS_SQL.format(api_information_url=api_information_url))
    if len(results) == 0:
        return []

    returns_list = []
    if not start_at_second and not prior_list_found:
        returns_list = get_first_list(api_information_url)
    elif start_at_second and not prior_list_found:
        returns_list = get_api_second_list(api_information_url)
    elif start_at_second and prior_list_found:
        returns_list = get_api_third_list(api_information_url)

    return returns_list



from shopagent.models import Mutation, MutationArgument, MutationReturn

from .arguments import get_api_arguments
from .code_examples import get_api_examples
from .description import get_api_description
from .fields_and_connections import get_api_fields_or_connections
from .returns import get_api_returns


def get_mutation_info(label, api_information_url):
    name = label
    description = get_api_description(api_information_url)
    arguments = get_api_arguments(api_information_url)
    returns = get_api_returns(api_information_url, False, len(arguments) > 0)
    examples = get_api_examples(api_information_url)

    mutation = Mutation(
        name=name,
        description=description.strip(),
        arguments=[MutationArgument(
            name=argument[0],
            type=argument[1],
            required=len(argument[2]) > 0,
            description=argument[3].strip(),
        ) for argument in arguments],
        returns=[MutationReturn(
            name=return_[0],
            type=return_[1],
            description=return_[2].strip()
        ) for return_ in returns],
        examples=examples
    )

    return mutation



from shopagent.models import Query, QueryArgument, QueryReturn

from .arguments import get_api_arguments
from .code_examples import get_api_examples
from .description import get_api_description
from .returns import get_api_returns


def get_query_info(label, api_information_url):
    name = label
    description = get_api_description(api_information_url)
    arguments = get_api_arguments(api_information_url)
    returns = get_api_returns(api_information_url, False, len(arguments) > 0)
    examples = get_api_examples(api_information_url)

    query = Query(
        name=name,
        description=description.strip(),
        arguments=[QueryArgument(
            name=argument[0],
            type=argument[1],
            required=len(argument[2]) > 0,
            description=argument[3].strip()
        ) for argument in arguments],
        returns=[QueryReturn(
            name=r[0],
            type=r[1],
            description=r[2].strip(),
        ) for r in returns],
        examples=examples,
    )

    return query



import json
import os.path

from ollama import chat
from whoosh import index
from whoosh.analysis import StemmingAnalyzer
from whoosh.fields import Schema, TEXT
from whoosh.qparser import QueryParser

from shopagent.constants import OLLAMA_MODEL
from shopagent.models import ShopifyAPI


def establish_agent_gen_prompt():
    return {
        'role': 'user',
        'content': """You are a mid-level software engineer who writes "Shopify agents" in Ruby using the Shopify GraphQL API.

Based on the provided request, use the [search_shopify_graphql] tool to obtain information about the relevant GraphQL functionalities to use. When you're ready to write the agent, use the [define_agent] tool to do so."""
    }


def format_user_request_into_prompt(user_request):
    # If does not already end with punctuation
    if user_request[-1] not in [".", "?", "!"]:
        user_request += "."

    return {
        'role': 'user',
        'content': f"Request: {user_request}",
    }


search_shopify_graphql_tool = {
    'type': 'function',
    'function': {
        'name': 'search_shopify_graphql',
        'description': 'Given a query, search through the Shopify GraphQL spec for one of "MUTATION", "QUERY", or "OBJECT". May be helpful for getting information if something is not already known beforehand',
        'parameters': {
            'type': 'object',
            'required': ['query', 'search_type'],
            'properties': {
                'query': {
                    'type': 'string',
                    'description': 'The query to search the Shopify GraphQL spec with',
                },
                'search_type': {
                    'type': 'string',
                    'enum': ['MUTATION', 'QUERY', 'OBJECT'],
                },
            }
        }
    },
}


def search_shopify_graphql(query, search_type):
    with open("shopify_api.json", "r") as shopify_api_file:
        data = json.load(shopify_api_file)
        api_information = ShopifyAPI(**data)

        if not os.path.exists("indexdir"):
            os.mkdir("indexdir")

        schema = Schema(
            name=TEXT(stored=True),
            description=TEXT(analyzer=StemmingAnalyzer()),
            fields_and_connections=TEXT(stored=True),
            arguments=TEXT(stored=True),
            returns=TEXT(stored=True),
            code_examples=TEXT(stored=True),
        )

        ix = index.create_in("indexdir", schema)
        writer = ix.writer()
        for category in api_information.categories:
            if search_type == "MUTATION":
                for mutation in category.mutations:
                    writer.add_document(
                        name=mutation.name,
                        description=mutation.description,
                        arguments=f"[{', '.join(['<' + arg.name + ': ' + arg.type + ' (required)' if arg.required else '' + '>' for arg in mutation.arguments])}]",
                        returns=f"[{', '.join(['<' + ret.name + ': ' + ret.type for ret in mutation.returns])}]",
                        code_examples='\n\n'.join(example.description + ':\n' + example.code for example in mutation.examples),
                    )
            elif search_type == "QUERY":
                for query in category.queries:
                    writer.add_document(
                        name=query.name,
                        description=query.description,
                        arguments=f"[{', '.join(['<' + arg.name + ': ' + arg.type + ' (required)' if arg.required else '' + '>' for arg in query.arguments])}]",
                        returns=f"[{', '.join(['<' + ret.name + ': ' + ret.type for ret in query.returns])}]",
                        code_examples='\n\n'.join(example.description + ':\n' + example.code for example in query.examples),
                    )
            else:
                for obj in category.objects:
                    writer.add_document(
                        name=obj.name,
                        description=obj.description,
                        fields_and_connections=f"[{', '.join(['<' + field.name + ': ' + field.type + '>' for field in obj.fields_and_connections])}]",
                        code_examples='\n\n'.join(example.description + ':\n' + example.code for example in obj.examples),
                    )

        qp = QueryParser("description", schema=ix.schema)
        q = qp.parse(query)

        with ix.searcher() as searcher:
            results = searcher.search(q, limit=10)
            return str([r for r in list(results)])


define_agent_tool = {
    'type': 'function',
    'function': {
        'name': 'define_agent',
        'description': 'Given a string containing a ruby program that defines a "Shopify agent", write to the file "agent.rb"',
        'parameters': {
            'type': 'object',
            'required': ['ruby_code'],
            'properties': {
                'query': {
                    'type': 'string',
                    'description': 'The query to search the Shopify GraphQL spec with',
                },
                'search_type': {
                    'type': 'string',
                    'enum': ['MUTATION', 'QUERY', 'OBJECT'],
                },
            }
        }
    },
}


def define_agent(ruby_code=''):
    if len(ruby_code) == 0:
        return "Cannot define an empty agent"

    with open("agent.rb", "w") as agent_file:
        agent_file.write(ruby_code)
        return "Defined! No more need to call tools"


def generate_agent(user_request):
    last_tool_calls = [[]]
    messages = [establish_agent_gen_prompt(), format_user_request_into_prompt(user_request)]
    available_functions = {
        'search_shopify_graphql': search_shopify_graphql,
        'define_agent': define_agent,
    }
    available_functions_schemas = {
        'search_shopify_graphql': search_shopify_graphql_tool,
        'define_agent': define_agent_tool,
    }

    defined_agent = False
    while not defined_agent:
        if len(messages) > 24:
            messages = [establish_agent_gen_prompt(), format_user_request_into_prompt(user_request)]

        response = chat(
            OLLAMA_MODEL,
            messages=messages,
            tools=[search_shopify_graphql_tool, define_agent_tool],
        )

        if response.message.tool_calls:
            print("Calling a tool...")
            # There may be multiple tool calls in the response
            for tool in response.message.tool_calls:
                # Ensure the function is available, and then call it
                if function_to_call := available_functions.get(tool.function.name):
                    try:
                        output = function_to_call(**tool.function.arguments)
                        if tool.function.name == "define_agent" and output != "Cannot define an empty agent":
                            defined_agent = True
                    except Exception as e:
                        output = f"You called the tool incorrectly, here's the error:\n{str(e)}\n\nThe expected parameters are [{', '.join(k for k in available_functions_schemas[tool.function.name]['function']['parameters']['properties'])}]"
                    last_tool_calls += [[]]

                # Only needed to chat with the model using the tool call results
                if response.message.tool_calls:
                    # Add the function response to messages for the model to use
                    messages.append(response.message)
                    messages.append({'role': 'tool', 'content': str(output), 'name': tool.function.name})
        elif "```ruby" in response.message.content and not defined_agent:
            ruby_code = response.message.content[response.message.content.index("```ruby")+7:]
            ruby_code = ruby_code[:ruby_code.index("```")]
            if len(ruby_code) > 0:
                define_agent(ruby_code)
                defined_agent = True
        elif "```" in response.message.content and not defined_agent:
            messages.append(response.message)
            messages.append({"role": "user", "content": "Use the [define_agent] tool instead of providing code"})
            last_tool_calls += [[]]


def gen_agent():
    user_request = input("What would you like your agent to do? ")
    generate_agent(user_request)



import os
import psycopg2
from time import sleep


GLOBAL_CONN = None
def establish_connection():
    global GLOBAL_CONN

    if GLOBAL_CONN is None:
        while True:
            try:
                GLOBAL_CONN = psycopg2.connect(
                    host="lsd.so",
                    database=os.environ.get("LSD_USER"),
                    user=os.environ.get("LSD_USER"),
                    password=os.environ.get("LSD_API_KEY"),
                    port="5432",
                )
                break
            except Exception as e:
                print("Ran into an issue connecting (now sleeping before trying again):", e)
                sleep(1)

    # Try out a simple request before handing back the postgres connection
    # in case, say, the wifi suddenly cut out after having already been
    # running the script a while. Yeah, not fun
    try:
        with GLOBAL_CONN.cursor() as curs:
            curs.execute("FROM https://lsd.so |> SELECT title")
            rows = curs.fetchall()
    except Exception as e:
        GLOBAL_CONN = None
        return establish_connection()

    return GLOBAL_CONN


def run_lsd(lsd_sql, retrying=False):
    conn = establish_connection()
    try:
        with conn.cursor() as curs:
            curs.execute(lsd_sql)
            rows = curs.fetchall()
            return [list(r) for r in rows]
    except Exception as e:
        if retrying:
            return []

        global GLOBAL_CONN
        GLOBAL_CONN = None
        return run_lsd(lsd_sql, True)



import re
from time import sleep

from shopagent.constants import EXAMPLE_LANGUAGE
from shopagent.lsd import run_lsd
from shopagent.models import CodeExample


CODE_EXAMPLES_EXISTENCE_SQL = """
PORTAL <| OPEN |

container <| ul[class*="Examples"] li |
item <| li |

FROM {api_information_url}
|> GROUP BY container
|> SELECT item
"""

CODE_EXAMPLE_SQL = """
PORTAL <| OPEN |

example_code_container <| div[class*="Examples"] |
example_code <| code[title="{language}"] |

FROM {modified_url}
|> GROUP BY example_code_container
|> SELECT example_code
"""


def get_api_example_code(api_information_url, example_description, retrying=False):
    simplified_language = EXAMPLE_LANGUAGE.lower() if "." not in EXAMPLE_LANGUAGE else EXAMPLE_LANGUAGE[:EXAMPLE_LANGUAGE.index(".")].lower()
    simplified_example = re.sub(r'[^a-zA-Z0-9\-]', '', example_description.replace(" ", "-").lower())
    modified_url = f"{api_information_url}?language={simplified_language}&example={simplified_example}"

    try:
        actual_example = run_lsd(CODE_EXAMPLE_SQL.format(language=EXAMPLE_LANGUAGE, modified_url=modified_url))
    except Exception as e:
        # Likely no rows returned from this page
        return ''

    if len(actual_example) == 0 or len(actual_example[0]) == 0:
        return ''

    if len(actual_example[0]) > 0 and ('Failed to obtain new columns' in actual_example[0][0] or 'Failed to [FROM]' in actual_example[0][0]):
        if retrying:
            return ''
        sleep(1)
        return get_api_example_code(api_information_url, example_description, True)

    return actual_example[0][0] if type(actual_example[0]) is list else actual_example[0]

def get_available_examples(api_information_url):
    available_examples = run_lsd(CODE_EXAMPLES_EXISTENCE_SQL.format(api_information_url=api_information_url))

    if len(available_examples) == 0:
        return []

    if type(available_examples[0]) is list:
        available_examples = [item for row in available_examples for item in row]

    if 'Failed to obtain new columns' in available_examples[0]:
        sleep(1)
        return get_available_examples(api_information_url)

    return available_examples

def get_api_examples(api_information_url):
    available_examples = get_available_examples(api_information_url)

    if len(available_examples) == 0:
        return []

    if type(available_examples[0]) is list:
        available_examples = [item for row in available_examples for item in row]

    api_examples = []
    for code_example in available_examples:
        actual_code = get_api_example_code(api_information_url, code_example)
        if len(actual_code) > 0:
            api_examples += [CodeExample(
                description=code_example.strip(),
                code=actual_code,
            )]

    return api_examples



from shopagent.lsd import run_lsd
from shopagent.models import ObjectFieldOrConnection

from .first_list import get_first_list
from .second_list import get_api_second_list


HAS_API_FIELDS = """
api_fields <| a#fields |

FROM {api_information_url}
|> GROUP BY api_fields
|> SELECT api_fields
"""

HAS_API_FIELDS_AND_CONNECTIONS = """
api_fields_and_connections <| a#fieldsandconnections |

FROM {api_information_url}
|> GROUP BY api_fields_and_connections
|> SELECT api_fields_and_connections
"""

def get_api_fields_or_connections(api_information_url, should_get_first_list=False):
    global HAS_API_FIELDS
    global HAS_API_FIELDS_AND_CONNECTIONS

    results = run_lsd(HAS_API_FIELDS.format(api_information_url=api_information_url))
    fields_or_connections_list = []
    if len(results) > 0:
        fields_or_connections_list = get_first_list(api_information_url) if should_get_first_list else get_api_second_list(api_information_url)

    results = run_lsd(HAS_API_FIELDS_AND_CONNECTIONS.format(api_information_url=api_information_url))
    if len(results) > 0:
        fields_or_connections_list = get_first_list(api_information_url) if should_get_first_list else get_api_second_list(api_information_url)

    if len(fields_or_connections_list) == 0:
        return []

    fields_or_connections = []
    for field_or_connection in fields_or_connections_list:
        name = field_or_connection[0]
        description = field_or_connection[3]
        field_type = field_or_connection[1]
        not_nullable = len(field_or_connection[2]) > 0

        fields_or_connections += [ObjectFieldOrConnection(
            name=name,
            type=field_type,
            description=description.strip(),
            not_nullable=not_nullable,
        )]

    return fields_or_connections



import json
import os.path

from requests import get
from rich import print

from shopagent.models import ShopifyAPI, ShopifyNavResponse
from shopagent.constants import SHOPIFY_BASE_URL, SHOPIFY_NAV_API_URL


def complete_nav_url(href):
    return f"{SHOPIFY_BASE_URL}{href}"


def get_top_level_operations():
    res = get(SHOPIFY_NAV_API_URL).json()
    shopify_nav = ShopifyNavResponse(**res)
    return shopify_nav


def print_data():
    if not os.path.exists("shopify_api.json"):
        sys.exit("Missing [shopify_api.json] file")
        return

    with open("shopify_api.json", "r") as shopify_api_file:
        data = json.load(shopify_api_file)
        api_information = ShopifyAPI(**data)

        print(api_information)



from shopagent.lsd import run_lsd


API_DESCRIPTION_SQL = """
description <| div.text > div.markdown |

FROM {api_information_url}
|> SELECT description
"""


def get_api_description(api_information_url):
    global API_DESCRIPTION_SQL

    results = run_lsd(API_DESCRIPTION_SQL.format(api_information_url=api_information_url))
    return results[0][0] if len(results) > 0 else ""



from shopagent.lsd import run_lsd

from .first_list import get_first_list


HAS_API_ARGUMENTS_SQL = """
arguments_anchor <| a#arguments |

FROM {api_information_url}
|> GROUP BY arguments_anchor
|> SELECT arguments_anchor
"""


def get_api_arguments(api_information_url):
    global HAS_API_ARGUMENTS_SQL

    results = run_lsd(HAS_API_ARGUMENTS_SQL.format(api_information_url=api_information_url))
    if len(results) == 0:
        return []

    return get_first_list(api_information_url)



from shopagent.lsd import run_lsd


GET_LIST_OF_ITEMS_SQL = """
list_item <| div[data-testid="primary-column"] > div:nth-of-type({n}) > div[data-testid="gql-list"]:first-of-type dl > div[class*="GqlListItem"] |
list_item_name <| dt .group\/hashtarget > span |
list_item_type <| div > a.underline |
list_item_non_nullable <| span[class*="NonNullChip"]! |
list_item_description <| dd div.markdown |


FROM {api_information_url}
|> GROUP BY list_item
|> SELECT list_item_name, list_item_type, list_item_non_nullable, list_item_description
"""


def get_list_of_items(api_information_url, n):
    global GET_LIST_OF_ITEMS_SQL
    return run_lsd(GET_LIST_OF_ITEMS_SQL.format(api_information_url=api_information_url, n=str(n)))



from shopagent.api.api import get_api_information, get_data
from shopagent.api.code_examples import get_api_examples
from shopagent.api.list import get_list_of_items
from shopagent.api.mutation import get_mutation_info
from shopagent.api.object import get_object_info
from shopagent.api.query import get_query_info



from shopagent.lsd import run_lsd

from .list import get_list_of_items


def get_first_list(api_information_url):
    counter = 0
    first_list = []
    while len(first_list) == 0:
        first_list = get_list_of_items(api_information_url, counter)
        counter += 1

    return first_list



from setuptools import setup

setup(
    name="shopagent",
    version="0.7.0",
    packages=['shopagent', 'shopagent.api', 'shopagent.code_gen', 'shopagent.constants', 'shopagent.lsd', 'shopagent.nav']
)
